
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/malloc.h>

#include "ods2.h"

/*
 * Convert a component of a pathname into a pointer to a locked inode.
 * This is a very central and rather complicated routine.
 * If the file system is not maintained in a strict tree hierarchy,
 * this can result in a deadlock situation (see comments in code below).
 *
 * The flag argument is LOOKUP, CREATE, RENAME, or DELETE depending on
 * whether the name is to be looked up, created, renamed, or deleted.
 * When CREATE, RENAME, or DELETE is specified, information usable in
 * creating, renaming, or deleting a directory entry may be calculated.
 * If flag has LOCKPARENT or'ed into it and the target of the pathname
 * exists, lookup returns both the target and its parent directory locked.
 * When creating or renaming and LOCKPARENT is specified, the target may
 * not be ".".  When deleting and LOCKPARENT is specified, the target may
 * be "."., but the caller must check to ensure it does an vrele and iput
 * instead of two iputs.
 *
 * Overall outline of ufs_lookup:
 *
 *	search for name in directory, to found or notfound
 * notfound:
 *	if creating, return locked directory, leaving info on available slots
 *	else return error
 * found:
 *	if at end of path and deleting, return information to allow delete
 *	if at end of path and rewriting (RENAME and LOCKPARENT), lock target
 *	  inode and return info to allow rewrite
 *	if not at end, add name to cache; if at end and neither creating
 *	  nor deleting, add name to cache
 *
 * NOTE: (LOOKUP | LOCKPARENT) currently returns the parent inode unlocked.
 */

int integer_parse __P(( char *name, int len ));

void make_version_string __P(( unsigned short int version,
                               char *version_string,
                               unsigned int *version_length ));
int search_dir_record __P(( struct mount *mp,
                            unsigned mntflags,
                            unsigned char *buffer,
                            char *name,
                            int len,
                            ino_t *ino ));
int search_dir_block __P(( struct mount *mp,
                           unsigned mntflags,
                           unsigned char *buffer,
                           char *name,
                           int len,
                           ino_t *ino ));

void make_version_string( unsigned short int version,
                          char *version_string,
                          unsigned int *version_length )
{
    int divisor;
    char *p = version_string;

    *version_length = 0;

    if (version >= 10000)      divisor = 10000;
    else if (version >=  1000) divisor =  1000;
    else if (version >=   100) divisor =   100;
    else if (version >=    10) divisor =    10;
    else                       divisor =     1;

    while (divisor > 0)
    {
        *p++ = (version / divisor) + '0';
        version %= divisor;
        divisor /= 10;
        (*version_length)++;
    }
}

int search_dir_record( struct mount *mp,
                       unsigned mntflags,
                       unsigned char *buffer,
                       char *name,
                       int len,
                       ino_t *ino )
{
    char vms_name[86];      /* MAX VMS NAME LENGTH */ /* XXX symbolic please */
    char version_string[5]; /* MAX VMS VERSION LENGTH */
    int status;
    unsigned pad_count;
    unsigned namestring;
    unsigned number_bytes;
    unsigned overhead;
    unsigned char *p = buffer;
    unsigned char *versionp;
    unsigned int i;
    unsigned version_length;
    unsigned full_length;
    unsigned compare_length;
    struct file_header_t *fh2;
    struct dir_record_t *dir_record;
    struct dir_entry_t *dir_entry;

    dir_record = (struct dir_record_t *) buffer;

    pad_count = dir_record->dir_b_namecount % 2;
    namestring = dir_record->dir_b_namecount + pad_count;

    /* From Kirby: */

    overhead = DIR_S_DIR_RECORD_OVERHEAD;
    number_bytes = ( dir_record->dir_w_size - overhead - namestring );
    p = dir_record->dir_t_name + namestring;

    bcopy(dir_record->dir_t_name, vms_name, dir_record->dir_b_namecount);

    versionp = vms_name + dir_record->dir_b_namecount;
    bcopy(".", versionp, 1); /* . instead of ';' */
    versionp++;

    for (i=0; i < number_bytes; )
    {
        ino_t look_ino;
        unsigned extension = 1;
        dir_entry = (struct dir_entry_t *) p;

        look_ino = FILE_NUMBER(dir_entry->dir_w_fid);

        if ((mntflags & ODS2MNT_FAST_NODIREXT) &&
            (dir_entry->dir_w_version == 1))
        {
            char *p = vms_name + dir_record->dir_b_namecount - 4;
            extension = bcmp(p, ".DIR", 4);
        }
        else if ((mntflags & ODS2MNT_NODIREXT) &&
                 (dir_entry->dir_w_version == 1))
        {
            /*
             * If the extension is NOT .DIR;1, then don't bother with the
             * file header.
             */

            char *p = vms_name + dir_record->dir_b_namecount - 4;
            unsigned not_dirext = bcmp(p, ".DIR", 4);

            if (!not_dirext)
            {
                status = ods2_read_file_header( mp, look_ino, &fh2 );
                if (status) return 1;

                extension = !( fh2->fh2_l_filechar & FCH_M_DIRECTORY );

                FREE( fh2, M_ODS2FSFH );
            }
        }

        if (extension)
        {
            make_version_string(dir_entry->dir_w_version,
                                version_string,
                                &version_length);

            bcopy(version_string, versionp, version_length);

            full_length = dir_record->dir_b_namecount + 1 + version_length;
        }
        else
            full_length = dir_record->dir_b_namecount - 1 - 3;

        if (len == full_length)
        {
            compare_length = len < full_length ? len : full_length;

            status = bcmp(name, vms_name, compare_length);
            if ( status == 0 )
            {
                *ino = look_ino;
                return 0;
            }
        }

        p += DIR_S_ENTRY_FID;
        i += DIR_S_ENTRY_FID;
    }

    return 1;
}

int search_dir_block( struct mount *mp,
                      unsigned mntflags,
                      unsigned char *buffer,
                      char *name,
                      int len,
                      ino_t *ino )
{
    int i;
    int status;
    int compare_length;
    short int *record_length;
    unsigned char *p = buffer;
    int end_of_block;
    unsigned length_maybe;
    struct dir_record_t *dir_record;

    record_length = (short int *) p;
    end_of_block = *record_length == -1;

    while (!end_of_block)
    {
        dir_record = (struct dir_record_t *) p;

        compare_length = 0;

if (0)
{
char xxname[255];
memcpy(xxname, dir_record->dir_t_name, dir_record->dir_b_namecount);
xxname[dir_record->dir_b_namecount] = '\0';

printf("looking for %s, found %s\n", name, xxname);
printf("Length of record=%d\n",
       (dir_record->dir_w_size + sizeof(dir_record->dir_w_size)));
}

        if (name[0] < dir_record->dir_t_name[0])
            return -1;

        length_maybe = len;
        if ((mntflags & ODS2MNT_FAST_NODIREXT) ||
            (mntflags & ODS2MNT_NODIREXT))
             length_maybe += 4;

        if (length_maybe >= dir_record->dir_b_namecount)
        {
            compare_length = dir_record->dir_b_namecount < len ?
                             dir_record->dir_b_namecount : len;

            status = 0;

            for (i=0; i < compare_length && status == 0; i++)
                status = name[i] - dir_record->dir_t_name[i];

            if ( status < 0 )
                return status;
            else if ( status == 0 )
            {
                status = search_dir_record( mp, mntflags, p, name, len, ino );
                if ( status == 0 ) return 0;
            }
        }

        p += dir_record->dir_w_size + sizeof(dir_record->dir_w_size);
        record_length = (short int *) p;
        end_of_block = *record_length == -1;
    }

    return 1;
}

int integer_parse( char *value, int len )
{
    int i;
    int sum = 0;

    for ( i=0; i < len; i++ )
        if ( value[i] >= '0' && value[i] <= '9' )
            sum = (sum * 10) + value[i] - '0';
        else
            return -1;

    return sum;
}

int
ods2_lookup(ap)
	struct vop_cachedlookup_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
	} */ *ap;
{
	register struct vnode *vdp;	/* vnode for directory being searched */
	register struct ods2_node *dp;	/* inode for directory being searched */
	register struct ods2_mount *imp;/* file system that directory is in */
	int lockparent;			/* 1 => lockparent flag is set */
	int wantparent;			/* 1 => wantparent or lockparent flag */
	int error;
	ino_t ino = 0;
	int len;
	char *name;
	struct vnode **vpp = ap->a_vpp;
	struct componentname *cnp = ap->a_cnp;
	int flags = cnp->cn_flags;
        int status;
        unsigned lbn;
	struct proc *p = cnp->cn_proc;

    unsigned lowercase;
    unsigned mntflags;
    char look_for[ 86 ];               /* Really only need 86 */

    unsigned char block[ODS2_BLOCK_SIZE];

    unsigned low;
    unsigned high;
    unsigned done;

    if (0)
        printf("ods2_lookup[starting]\n");

	*vpp = NULL;
	vdp = ap->a_dvp;
	dp = VTOODS2(vdp);
	imp = MOUNTTOODS2MOUNT(dp->mount);
        lowercase = imp->flags & ODS2MNT_LOWERCASE;
        mntflags = imp->flags;
	lockparent = flags & LOCKPARENT;
	wantparent = flags & (LOCKPARENT|WANTPARENT);

	/*
	 * We now have a segment name to search for, and a directory to search.
	 */

	len = cnp->cn_namelen;
	name = cnp->cn_nameptr;

    bcopy(name, look_for, len);
    if (lowercase)
    {
        /*
         * Upcase the name we're looking for.  Why?
         *   1.  CPUTBL.LIS < CPU_COMMON.LIS, but
         *       cputbl.lis > cpu_common.lis
         *   2.  We upcase one string only, instead of lowercasing many names.
         */

        int i;

        for (i=0; i < len; i++)
            if (look_for[i] >= 'a' && look_for[i] <= 'z')
                look_for[i] -= ' ';
    }

    if (0)
        printf("ods2_lookup[\tlooking for\n\t%s\n\t%s\n\tflags=%x]\n",
               name, look_for, flags);

    if ( len == 1 && !bcmp(look_for, ".", len) )
    {
        VREF( vdp );
        *vpp = vdp;

        if (cnp->cn_flags & MAKEENTRY)
            cache_enter(vdp, *vpp, cnp);

        return 0;
    }
    else if ( len == 2 && !bcmp(look_for, "..", len) )
    {
        VOP_UNLOCK(vdp, 0, p);  /* race to get the inode */

        ino = FILE_NUMBER(dp->file_header->fh2_w_backlink);
        error = ods2_vget( vdp->v_mount, ino, vpp );

        if (error) {
            vn_lock(vdp, LK_EXCLUSIVE | LK_RETRY, p);
            return (error);
        }
        if (lockparent && (flags & ISLASTCN) &&
            (error = vn_lock(vdp, LK_EXCLUSIVE, p)))
        {
            vput(*vpp);
            return (error);
        }

        return 0;
    }
    else if ( len == 1 &&
              dp->i_num == MFD_FILE_NUMBER &&
              !bcmp(look_for, "#", len) )
    {
        error = ods2_vget( dp->mount, imp->fid_directory_fid, vpp );
        if ( error ) return error;

        if (cnp->cn_flags & MAKEENTRY)
            cache_enter(vdp, *vpp, cnp);

        if (!lockparent)
            VOP_UNLOCK(vdp, 0, p);

        return 0;
    }

    if ( dp->i_num == imp->fid_directory_fid )
    {
        /*
         * The name should be an integer which is the file number of the file.
         */

        ino = integer_parse( look_for, len );

if (0)
printf("ods2_lookup[looking in #, number conversion of %s = %d\n",
       look_for, ino );

        if ( ino <= 0 || ino > imp->maxfiles )
        {
            if ( !lockparent )
                VOP_UNLOCK(vdp, 0, p);

            return ENOENT;
        }

        /*
         * Perform a vget on the inode number.  It may or may not be there
         * depending if the file was on the surface to start with.
         */

        status = ods2_vget( dp->mount, ino, vpp );
        if ( status )
            return status;

        if ( !lockparent )
            VOP_UNLOCK(vdp, 0, p);

        if (cnp->cn_flags & MAKEENTRY)
            cache_enter(vdp, *vpp, cnp);

        return status;
    }

    low = 1;
    high = ( dp->i_size - ODS2_BLOCK_SIZE ) / ODS2_BLOCK_SIZE;

    do
    {
      unsigned int mid = (high + low) / 2;

      error = VOP_BMAP(vdp, mid, NULL, &lbn, NULL, NULL);
      if (error) return error;

      if (0)
          printf("ods2_lookup[reading directory block VBN # %d, LBN # %d]\n",
                 mid, lbn);

      error = ods2_bread(imp->ods2_devvp, &imp->geometry, lbn, NOCRED, block);
      if (error)
          return error;

      status = search_dir_block( dp->mount,
                                 mntflags,
                                 block,
                                 look_for,
                                 len,
                                 &ino );

      if ( status == 0 )
      {
          if (!lockparent)
              VOP_UNLOCK(vdp, 0, p);

          status = ods2_vget( dp->mount, ino, vpp );
          if ( status )
              return status;

          if (cnp->cn_flags & MAKEENTRY)
              cache_enter(vdp, *vpp, cnp);

          return status;
      }
      else if ( status < 0 )
          high = mid - 1;
      else
          low = mid + 1;

      done = low > high;
    } while (!done);

    if (!lockparent)
        VOP_UNLOCK(vdp, 0, p);

if (0)
    printf( "ods2_lookup[did not find %s : %s]\n", name, look_for );

    return ENOENT;
}
